/* this file is autogenerated -- do not edit */
/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef ___SCREENSHOOTER_MARSHAL_MARSHAL_H__
#define ___SCREENSHOOTER_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:STRING (./lib/screenshooter-marshal.list:1) */
#define _screenshooter_marshal_VOID__STRING	g_cclosure_marshal_VOID__STRING

/* VOID:POINTER,STRING (./lib/screenshooter-marshal.list:2) */
extern
void _screenshooter_marshal_VOID__POINTER_STRING (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:STRING,STRING (./lib/screenshooter-marshal.list:3) */
extern
void _screenshooter_marshal_VOID__STRING_STRING (GClosure     *closure,
                                                 GValue       *return_value,
                                                 guint         n_param_values,
                                                 const GValue *param_values,
                                                 gpointer      invocation_hint,
                                                 gpointer      marshal_data);


G_END_DECLS

#endif /* ___SCREENSHOOTER_MARSHAL_MARSHAL_H__ */
